@@ -12,17 +12,16 @@ import Control.Monad.Trans.Control
 
 -- | Use lazy I\/O to consume all elements from a @Source@.
 --
--- Since 0.0.0
+-- Since 0.2.0
 lazyConsume :: MonadBaseControl IO m => Source m a -> ResourceT m [a]
-lazyConsume (Source msrc) = do
-    src <- msrc
-    go src
+lazyConsume src0 = do
+    go src0
   where
 
     go src = liftBaseOp_ unsafeInterleaveIO $ do
         res <- sourcePull src
         case res of
             Closed -> return []
-            Open x -> do
-                y <- go src
+            Open src' x -> do
+                y <- go src'
                 return $ x : y
