@@ -28,12 +28,14 @@ module Data.Conduit.List
       -- ** Pure
     , map
     , concatMap
+    , concatMapAccum
     , groupBy
     , isolate
     , filter
       -- ** Monadic
     , mapM
     , concatMapM
+    , concatMapAccumM
     ) where
 
 import Prelude
@@ -49,19 +51,19 @@ import Data.Monoid (mempty)
 
 -- | A strict left fold.
 --
--- Since 0.0.0
+-- Since 0.2.0
 fold :: Resource m
      => (b -> a -> b)
      -> b
      -> Sink a m b
 fold f accum0 = sinkState
     accum0
-    (\accum input -> return (f accum input, Processing))
+    (\accum input -> return (StateProcessing $ f accum input))
     return
 
 -- | A monadic strict left fold.
 --
--- Since 0.0.0
+-- Since 0.2.0
 foldM :: Resource m
       => (b -> a -> m b)
       -> b
@@ -70,29 +72,31 @@ foldM f accum0 = sinkState
     accum0
     (\accum input -> do
         accum' <- lift $ f accum input
-        return (accum', Processing)
+        return $ StateProcessing accum'
     )
     return
 
 -- | Apply the action to all values in the stream.
 --
--- Since 0.0.0
+-- Since 0.2.0
 mapM_ :: Resource m
       => (a -> m ())
       -> Sink a m ()
-mapM_ f = Sink $ return $ SinkData
-    (\input -> lift (f input) >> return Processing)
-    (return ())
+mapM_ f =
+    SinkData push close
+  where
+    push input = lift (f input) >> return (Processing push close)
+    close = return ()
 
 -- | Convert a list into a source.
 --
--- Since 0.0.0
+-- Since 0.2.0
 sourceList :: Resource m => [a] -> Source m a
 sourceList l0 =
     sourceState l0 go
   where
-    go [] = return ([], Closed)
-    go (x:xs) = return (xs, Open x)
+    go [] = return StateClosed
+    go (x:xs) = return $ StateOpen xs x
 
 -- | Ignore a certain number of values in the stream. This function is
 -- semantically equivalent to:
@@ -102,7 +106,7 @@ sourceList l0 =
 -- However, @drop@ is more efficient as it does not need to hold values in
 -- memory.
 --
--- Since 0.0.0
+-- Since 0.2.0
 drop :: Resource m
      => Int
      -> Sink a m ()
@@ -111,12 +115,12 @@ drop count0 = sinkState
     push
     close
   where
-    push 0 x = return (0, Done (Just x) ())
+    push 0 x = return $ StateDone (Just x) ()
     push count _ = do
         let count' = count - 1
-        return (count', if count' == 0
-                            then Done Nothing ()
-                            else Processing)
+        return $ if count' == 0
+            then StateDone Nothing ()
+            else StateProcessing count'
     close _ = return ()
 
 -- | Take some values from the stream and return as a list. If you want to
@@ -125,7 +129,7 @@ drop count0 = sinkState
 --
 -- > take i = isolate i =$ consume
 --
--- Since 0.0.0
+-- Since 0.2.0
 take :: Resource m
      => Int
      -> Sink a m [a]
@@ -134,22 +138,21 @@ take count0 = sinkState
     push
     close
   where
-    push (0, front) x = return ((0, front), Done (Just x) (front []))
+    push (0, front) x = return (StateDone (Just x) (front []))
     push (count, front) x = do
         let count' = count - 1
             front' = front . (x:)
-            res = if count' == 0
-                    then Done Nothing (front' [])
-                    else Processing
-        return ((count', front'), res)
+        return $ if count' == 0
+                    then StateDone Nothing (front' [])
+                    else StateProcessing (count', front')
     close (_, front) = return $ front []
 
 -- | Take a single value from the stream, if available.
 --
--- Since 0.0.0
+-- Since 0.2.0
 head :: Resource m => Sink a m (Maybe a)
 head =
-    Sink $ return $ SinkData push close
+    SinkData push close
   where
     push x = return $ Done Nothing (Just x)
     close = return Nothing
@@ -157,81 +160,109 @@ head =
 -- | Look at the next value in the stream, if available. This function will not
 -- change the state of the stream.
 --
--- Since 0.0.0
+-- Since 0.2.0
 peek :: Resource m => Sink a m (Maybe a)
 peek =
-    Sink $ return $ SinkData push close
+    SinkData push close
   where
     push x = return $ Done (Just x) (Just x)
     close = return Nothing
 
 -- | Apply a transformation to all values in a stream.
 --
--- Since 0.0.0
+-- Since 0.2.0
 map :: Monad m => (a -> b) -> Conduit a m b
-map f = Conduit $ return $ PreparedConduit
-    { conduitPush = return . Producing . return . f
-    , conduitClose = return []
-    }
+map f =
+    conduit
+  where
+    conduit = Conduit push close
+    push = return . Producing conduit . return . f
+    close = return []
 
 -- | Apply a monadic transformation to all values in a stream.
 --
 -- If you do not need the transformed values, and instead just want the monadic
 -- side-effects of running the action, see 'mapM_'.
 --
--- Since 0.0.0
+-- Since 0.2.0
 mapM :: Monad m => (a -> m b) -> Conduit a m b
-mapM f = Conduit $ return $ PreparedConduit
-    { conduitPush = fmap (Producing . return) . lift . f
-    , conduitClose = return []
-    }
+mapM f =
+    conduit
+  where
+    conduit = Conduit push close
+    push = fmap (Producing conduit . return) . lift . f
+    close = return []
 
 -- | Apply a transformation to all values in a stream, concatenating the output
 -- values.
 --
--- Since 0.0.0
+-- Since 0.2.0
 concatMap :: Monad m => (a -> [b]) -> Conduit a m b
-concatMap f = Conduit $ return $ PreparedConduit
-    { conduitPush = return . Producing . f
-    , conduitClose = return []
-    }
+concatMap f =
+    conduit
+  where
+    conduit = Conduit push close
+    push = return . Producing conduit . f
+    close = return []
 
 -- | Apply a monadic transformation to all values in a stream, concatenating
 -- the output values.
 --
--- Since 0.0.0
+-- Since 0.2.0
 concatMapM :: Monad m => (a -> m [b]) -> Conduit a m b
-concatMapM f = Conduit $ return $ PreparedConduit
-    { conduitPush = fmap Producing . lift . f
-    , conduitClose = return []
-    }
+concatMapM f =
+    conduit
+  where
+    conduit = Conduit push close
+    push = fmap (Producing conduit) . lift . f
+    close = return []
+
+-- | 'concatMap' with accumerator.
+--
+-- Since 0.2.0
+concatMapAccum :: Resource m => (a -> accum -> (accum, [b])) -> accum -> Conduit a m b
+concatMapAccum f accum = conduitState accum push close
+  where
+    push state input = let (state', result) = f input state
+                       in return $ StateProducing state' result
+    close _ = return []
+
+-- | 'concatMapM' with accumerator.
+--
+-- Since 0.2.0
+concatMapAccumM :: Resource m => (a -> accum -> m (accum, [b])) -> accum -> Conduit a m b
+concatMapAccumM f accum = conduitState accum push close
+  where
+    push state input = do (state', result) <- lift (f input state)
+                          return $ StateProducing state' result
+    close _ = return []
 
 -- | Consume all values from the stream and return as a list. Note that this
 -- will pull all values into memory. For a lazy variant, see
 -- "Data.Conduit.Lazy".
 --
--- Since 0.0.0
+-- Since 0.2.0
 consume :: Resource m => Sink a m [a]
 consume = sinkState
     id
-    (\front input -> return (front . (input :), Processing))
+    (\front input -> return (StateProcessing $ front . (input :)))
     (\front -> return $ front [])
 
 -- | Grouping input according to an equality function.
 --
--- Since 0.0.2
+-- Since 0.2.0
 groupBy :: Resource m => (a -> a -> Bool) -> Conduit a m [a]
 groupBy f = conduitState
     []
     push
     close
   where
-    push []      v = return ([v], Producing [])
+    push []      v = return $ StateProducing [v] []
     push s@(x:_) v =
       if f x v then
-        return (v:s, Producing [])
+        return $ StateProducing (v:s) []
       else
-        return ([v], Producing [s])
+        return $ StateProducing [v] [s]
     close s = return [s]
 
 -- | Ensure that the inner sink consumes no more than the given number of
@@ -246,7 +277,7 @@ groupBy f = conduitState
 -- >     someOtherSink
 -- >     ...
 --
--- Since 0.0.0
+-- Since 0.2.0
 isolate :: Resource m => Int -> Conduit a m a
 isolate count0 = conduitState
     count0
@@ -256,35 +287,38 @@ isolate count0 = conduitState
     close _ = return []
     push count x = do
         if count == 0
-            then return (count, Finished (Just x) [])
+            then return $ StateFinished (Just x) []
             else do
                 let count' = count - 1
-                return (count',
-                    if count' == 0
-                        then Finished Nothing [x]
-                        else Producing [x])
+                return $ if count' == 0
+                    then StateFinished Nothing [x]
+                    else StateProducing count' [x]
 
 -- | Keep only values in the stream passing a given predicate.
 --
--- Since 0.0.0
+-- Since 0.2.0
 filter :: Resource m => (a -> Bool) -> Conduit a m a
-filter f = Conduit $ return $ PreparedConduit
-    { conduitPush = return . Producing . Prelude.filter f . return
-    , conduitClose = return []
-    }
+filter f =
+    conduit
+  where
+    conduit = Conduit push close
+    push = return . Producing conduit . Prelude.filter f . return
+    close = return []
 
 -- | Ignore the remainder of values in the source. Particularly useful when
 -- combined with 'isolate'.
 --
--- Since 0.0.0
+-- Since 0.2.0
 sinkNull :: Resource m => Sink a m ()
-sinkNull = Sink $ return $ SinkData
-    (\_ -> return Processing)
-    (return ())
+sinkNull =
+    SinkData push close
+  where
+    push _ = return $ Processing push close
+    close = return ()
 
 -- | A source that returns nothing. Note that this is just a type-restricted
 -- synonym for 'mempty'.
 --
--- Since 0.0.4
+-- Since 0.2.0
 sourceNull :: Resource m => Source m a
 sourceNull = mempty
